#!/bin/bash

######################
#
# Nessus Offline Patch Checker
#
# nopc.sh - 0.4.2
#
# This is an offline patch checker using the Nessus nasls and a manually created KB file
#
# The following OS/Distros are potentially supported, * means to follow, ** means untested
#
# ####################### FreeBSD ** ###########################################
# ####################### RedHat Linux (+ Centos/Fedora *) #####################
# ####################### Mandrake * ###########################################
# ####################### SuSE * ###############################################
# ####################### Gentoo * #############################################
# ####################### Debian (+ Ubuntu) ####################################
# ####################### Slackware ** #########################################
# ####################### MacOS X ##############################################
# ####################### Solaris ##############################################
# ####################### AIX ##################################################
# ####################### HP-UX ** #############################################
# ####################### Cisco IOS/ASA ########################################

# 
# The output stage is still not ideal, there are two line here one quick, lists you the outdated packages,
# but not the missing security patch, the other is slower (~x5, take about about 1 min to parse a rpm file)
# but will show you the nasl name, so you get the security advisory name. You will though have to tailor this
# for each OS you test as the name format is different, so ...
#
# To add more check look at the plugin:
#		
#		ssh_get_info.nasl
#
# Also deafult nasls location set for osx so you might want to update if you use it alot

nessus_dir='/Library/Nessus/run/lib/nessus/plugins/'

#
# Copyright (C) 1/1/2012 - rid@portcullis-security.com 
#
#       * This program is free software; you can redistribute it and/or modify
#       * it under the terms of the GNU General Public License as published by
#       * the Free Software Foundation; either version 2 of the License, or
#       * (at your option) any later version.
#       *
#       * This program is distributed in the hope that it will be useful,
#       * but WITHOUT ANY WARRANTY; without even the implied warranty of
#       * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       * GNU General Public License for more details.
#       *
#       * You should have received a copy of the GNU General Public License
#       * along with this program; if not, write to the Free Software
#       * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.     
#
#####################

printSystems(){
	printf " 1 - Redhat\n"
	printf " 2 - OSX\n"
	printf " 3 - Debian\n"
	printf " 4 - Ubuntu\n"
	printf " 5 - Slackware *\n"
	printf " 6 - Solaris (Maybe !11)\n"
	printf " 7 - AIX\n"
	printf " 8 - HP-UX *\n"
	printf " 9 - FreeBSD *\n"
	printf " 10 - Cisco ASA/IOS\n"

	printf "\n * UNTESTED!!\n\n"
}

printCMD(){ # function to print command line running if we needed to ask any questions

	if [ "$alloncmd" -gt 0 ]
		then 
			cmd="[+] To run this in a script the command would be:\n\n$0"
			
			if [ ! -z "$cmdnasl" ]
				then
				cmd="$cmd -n '$nasl'"
			fi
			
			if [ ! -z "$cmddir" ]
				then
				cmd="$cmd -d '$nessus_dir'"
			fi
			
			cmd="$cmd -s"

			for arg in "$@"
				do
					cmd="$cmd '$arg'"
				done
				
			printf "$cmd\n\n"
	fi

}	

getSystem(){ # function to determine what kind of system we are patch parsing for.

	let alloncmd++
	printf "[+] What type of system have you got the patch output for?\n"
	printSystems
	read -p "Enter 1-10? " choice
	system="$choice"
	return
	
}

getFile(){ # function to get a file imput from user, forst arg is the instrauction for the user, second is the preproulated value from command line

	text="$1"
	file="$2"
	while [ ! -e "$file" ]
		do 
			let alloncmd++
			printf "[+] %b \n" "$text"
			read -p "[+] Enter Location of file: " file
		done
}

getString(){ # function to get a string imput from user, first arg is the instrauction for the user, second is the preproulated value from command line

	text="$1"
	string="$2"
	if [ -z "$string" ]
		then
			let alloncmd++
			printf "[+] %b \n" "$text"
			read -p "[+] Enter Text Requested: " string
	fi	
}

setRedHat(){ # Each of the OS has a call back function selected from a case. 
	
	# Need to get one file and one string.
	
	system=1 # system number not important
	system_name="Redhat" # system name not important
	nessus_filematch="redhat-RHSA" # the format of the nasls that this check needs. See find command later.
	
	patch_ins="Run '/bin/rpm -qa --qf '%{NAME}-%{VERSION}-%{RELEASE}|%{EPOCH}\\\n' > patchlist.txt'" # the command to run to get the patch info
	version_ins="Enter the Contents of /etc/redhat-release"  # the string to put in for the version info
	patchline="Host/RedHat/rpm-list=" # the KB header for patch info
	versionline="Host/RedHat/release=" # the KB header for version info
	patch="${args[0]}" # get the patch if passed via the command liine
	version="${args[1]}" # get the version if passed via the command liine
	getFile "$patch_ins" "$patch" # get patch from user input
	patch="$file" 
	
	getString "$version_ins" "$version"  # get version from user input
	version="$string"
	
	printCMD "$system" "$patch" "$version";
	
 	
	v=`echo "$version" | perl -pi -e 's/\n/\\\\n/g'` # remove newlines
	p=`cat "$patch" | perl -pi -e 's/\n/\\\\n/g'` # remove newlines

	# create the kb file
	echo "$fileheader$versionline$v" >> $kb
	echo "$fileheader$patchline$p" >> $kb
		
}

# all other callbacks are of the same format, some have more parsing and more agruments

setOSX(){
	
	# Need to get two files and one string.
	
	system_name="OS X"
	nessus_filematch="macosx"

	patch_ins="Run 'grep -A 1 displayName /Library/Receipts/InstallHistory.plist 2>/dev/null| grep string | sed 's/<string>\(.*\)<\/string>.*/\1/g'  | sed 's/^[      ]*//g'|tr  -d -c 'a-zA-Z0-9\n _-'|sort|uniq > patchlist.txt'"
	bom_ins="Run \"ls -1 /Library/Receipts/boms /private/var/db/receipts 2>/dev/null | grep '\.bom$'\" > bomlist.txt"
	version_ins="Enter OS X Version\n e.g. 10.6.8"
	
	patchline="Host/MacOSX/packages="
	bomline="Host/MacOSX/packages/boms="
	versionline="Host/MacOSX/Version=Mac OS X "
	
	
	installed="${args[0]}"
	boms="${args[1]}"
	version="${args[2]}"
	
	getFile "$patch_ins" "$installed"
	patch="$file"
	getFile "$bom_ins" "$boms"
	bom="$file"
	getString "$version_ins" "$version"
	version="$string"
	
	printCMD "$system" "$patch" "$bom" "$version"

	v=`echo "$version" | perl -pi -e 's/\n/\\\\n/g'`
	p=`cat "$patch" | perl -pi -e 's/\n/\\\\n/g'`
	b=`cat "$bom" | perl -pi -e 's/\n/\\\\n/g'`
	
	echo "$fileheader$versionline$v" >> $kb
	echo "$fileheader$bomline$b" >> $kb
	echo "$fileheader$patchline$p" >> $kb
			
}

setDebian(){
	
	# Need to get one file and one string.
	
	system=3
	system_name="Debian"
	nessus_filematch="debian_DSA"
	
	patch_ins="Run 'dpkg -l|cat > patchlist.txt'"
	version_ins="Enter the Contents of /etc/debian_version"
	patchline="Host/Debian/dpkg-l="
	versionline="Host/Debian/release="
	patch="${args[0]}"
	version="${args[1]}"
	getFile "$patch_ins" "$patch"
	patch="$file"

	getString "$version_ins" "$version"
	version="$string"
	
	printCMD "$system" "$patch" "$version" 
		
	v=`echo "$version" | perl -pi -e 's/\n/\\\\n/g'`
	p=`cat "$patch" | perl -pi -e 's/\n/\\\\n/g'`

	echo "$fileheader$versionline$v" >> $kb
	echo "$fileheader$patchline$p" >> $kb
		
}


setUbuntu(){
	
	# Need to get one file and one string.
	
	system=4
	system_name="Ubuntu"
	nessus_filematch="ubuntu_USN"
	
	patch_ins="Run 'dpkg -l|cat > patchlist.txt'"
	version_ins="Enter the Value of DISTRIB_RELEASE=() from /etc/lsb-release e.g. 11.10"
	patchline="Host/Debian/dpkg-l="
	versionline="Host/Ubuntu/release="
	patch="${args[0]}"
	version="${args[1]}"
	getFile "$patch_ins" "$patch"
	patch="$file"

	getString "$version_ins" "$version"
	version="$string"
	
	printCMD "$system" "$patch" "$version" 	
	
	v=`echo "$version" | perl -pi -e 's/\n/\\\\n/g'`
	p=`cat "$patch" | perl -pi -e 's/\n/\\\\n/g'`

	echo "$fileheader$versionline$v" >> $kb
	echo "$fileheader$patchline$p" >> $kb
		
}

setSlackware(){ #UNTESTED 
	
	# Need to get one file and one string.
	
	system=5
	system_name="Slackware"
	nessus_filematch="Slackware_SSA"
	
	patch_ins="Run 'ls -1 /var/log/packages > patchlist.txt'"
	version_ins="Enter the Contents of /etc/slackware-version"
	patchline="Host/Slackware/packages="
	versionline="Host/Slackware/release="
	patch="${args[0]}"
	version="${args[1]}"
	getFile "$patch_ins" "$patch"
	patch="$file"

	getString "$version_ins" "$version"
	version="$string"
	
	printCMD "$system" "$patch" "$version"
	 	
	v=`echo "$version" | perl -pi -e 's/^Slackware //' | perl -pi -e 's/\n/\\\\n/g'`
	p=`cat "$patch" | perl -pi -e 's/\n/\\\\n/g'`

	echo "$fileheader$versionline$v" >> $kb
	echo "$fileheader$patchline$p" >> $kb
		
}

setSolaris(){ #UNTESTED - UNWRITTEN
	
	# Need to get two files and one string.
	# Host/Solaris/Version=5.7\r\n
	# Solaris/Packages/Versions/SUNWxfb=7.0.0,REV=1998.08.11\r
	#  Host/Solaris/showrev=Hostname: havefun\r\nHostid: 80f9249b\r\nRelease: 5.7\r\nKernel architecture: sun4u\r\nApplication architecture: sparc\r\nHardware provider: Sun_Microsystems\r\nDomain: 6     5535.com\r\nKernel version: SunOS 5.7 Generic 106541-08 October 1999\r\n\r\nOpenWindows version: \r\nOpenWindows Version 3.6.1  2 September 1999\r\n\r\nPatch: 107648-09 Obsoletes:  Requires: 107078-17 Inc     ompatibles:  Packages: SUNWxwrtx, SUNWxwrtl, SUNWxwice, SUNWxwplt, SUNWxwicx, SUNWxwplx, SUNWxwinc, SUNWxwman, SUNWxwpmn, SUNWxwslb, SUNWxwslx\r\nPatch: 107656-05 Obsoletes:  Requires: 107648-06 Incompati     bles:  Packages: SUNWxwrtx, SUNWxwrtl, SUNWxwplt, SUNWxwplx, SUNWxwpmn, SUNWxwslb, SUNWxwslx\r\nPatch: 107430-01 Obsoletes:  Requires:  Incompatibles:  Packages: SUNWwsr\r\nPatch: 108029-02 Obsoletes:  Re     quires:  Incompatibles:  Packages: SUNWwsr\r\nPatch: 107063-01 Obsoletes:  Requires:  Incompatibles:  Packages: SUNWtleux\r\nPatch: 107437-03 Obsoletes:  Requires:  Incompatibles:  Packages: SUNWtiu8x, SU     NWtiu8\r\nPatch: 107316-01 Obsoletes:  Requires:  Incompatibles:  Packages: SUNWploc,
	# Host/Solaris/pkginfo=SMCbash                 (sparc) 4.1\r\nSMCgcc                 (sparc) 3.4.6\r\nSMCgzip                 (sparc) 1.4\r\nSMClgc346                 (sparc) 3.4.6\r\nSMCliconv                      (sparc) 1.9.2\r\nSMClintl                 (sparc) 3.4.0\r\nSUNW5xmft                 (sparc) 7.0,REV=1.0.6\r\nSUNWab2m                 (sparc) 2.00,REV=19980819\r\nSUNWaccr                      (sparc) 11.7.0,REV=1998.09.01.04.16\r\nSUNWaccu 
	
	
	# /usr/bin/pkginfo -x | awk '{ if ( NR % 2 ) { prev = $1 } else  { print prev" "$0  } }'
	# /usr/bin/showrev -a
	# /usr/bin/pkginfo -x | awk '{ if ( NR % 2 ) { prev = $1 } else  { print prev" "$0  } }' | awk '{print $1"="$3}'
	
	system=6
	system_name="Solaris"
	nessus_filematch="solaris9"
	
	pkg_ins="Run '/usr/bin/pkginfo -x | awk '{ if ( NR % 2 ) { prev = \$1 } else  { print prev\" \"\$0  } }' > pkginfo.txt'"
	revision_ins="Run '/usr/bin/showrev -a > revision.txt'"
	version_ins="Enter the Solaris Release e.g 5.7"
	pkgline="Host/Solaris/pkginfo="
	revisionline="Host/Solaris/showrev="
	versionline="Host/Solaris/Version="

	version="${args[0]}"
	pkg="${args[1]}"
	revision="${args[2]}"

	getString "$version_ins" "$version"
	version="$string"

	getFile "$pkg_ins" "$pkg"
	pkg="$file"
	
	getFile "$revision_ins" "$revision"
	revision="$file"
	
	printCMD "$system" "$version" "$pkg" "$revision"
	 	
	v=`echo "$version" | perl -pi -e 's/\n/\\\\n/g'`
	p=`cat "$pkg" | perl -pi -e 's/\n/\\\\n/g'`
	r=`cat "$revision" | perl -pi -e 's/\n/\\\\r\\\\n/g'`


	echo "$fileheader$versionline$v" >> $kb
	echo "$fileheader$pkgline$p" >> $kb
	echo "$fileheader$revisionline$r" >> $kb
	
	cat "$pkg" | awk '{print "1323253913 1 Solaris/Packages/Versions/"$1"="$3}' >> $kb	
		
}

setAIX(){ 
	
	# Need to get one file and two strings.
	
	system=7
	system_name="AIX"
	nessus_filematch="aix_"
	
	patch_ins="Run 'lslpp -Lc > patchlist.txt'"
	version_ins="Enter the AIX Relese e.g. 6.1"
	oslevel_ins="Enter the output of 'oslevel -r' e.g. 6100-04"
	
	patchline="Host/AIX/lslpp="
	versionline="Host/AIX/version="
	oslevelline="Host/AIX/oslevel="
	
	patch="${args[0]}"
	version="${args[1]}"
	oslevel="${args[2]}"

	getFile "$patch_ins" "$patch"
	patch="$file"

	getString "$version_ins" "$version"
	version="$string"
	
	getString "$oslevel_ins" "$oslevel"
	oslevel="$string"
	
	printCMD "$system" "$patch" "$version" "$oslevel"
 	
	v=`echo "$version" |perl -pi -e 's/\n/\\\\n/g'`
	o=`echo "$oslevel" |perl -pi -e 's/\n/\\\\n/g'`
	p=`cat "$patch" | perl -pi -e 's/\n/\\\\n/g'`

	echo "$fileheader$versionline$v" >> $kb
	echo "$fileheader$oslevelline$o" >>$kb
	echo "$fileheader$patchline$p" >> $kb
}

setHPUX(){ #UNTESTED 
	
	# Need to get one file and two strings.
	
	system=8
	system_name="HP-UX"
	nessus_filematch="hpux_"
	
	patch_ins="Run '/usr/sbin/swlist -l fileset -a revision > patchlist.txt'"
	version_ins="Enter the HP-UX Release"
	proc_ins="Enter the processor type, ia64 or parisc"
	
	patchline="Host/HP-UX/swlist="
	versionline="Host/HP-UX/version="
	procline="Host/HP-UX/processor="
	
	patch="${args[0]}"
	version="${args[1]}"
	proc="${args[2]}"

	getFile "$patch_ins" "$patch"
	patch="$file"

	getString "$version_ins" "$version"
	version="$string"
	
	getString "$proc_ins" "$proc"
	proc="$string"
	
	printCMD "$system" "$patch" "$version" "$proc"
 	
	v=`echo "$version" |perl -pi -e 's/\n/\\\\n/g'`
	o=`echo "$proc" |perl -pi -e 's/\n/\\\\n/g'`
	p=`cat "$patch" | perl -pi -e 's/\n/\\\\n/g'`

	echo "$fileheader$versionline$v" >> $kb
	echo "$fileheader$procline$o" >>$kb
	echo "$fileheader$patchline$p" >> $kb
}

setFreeBSD(){
	
	# Need to get one file and one string.
	
	system=9
	system_name="Free BSD"
	nessus_filematch="freebsd_"

	patch_ins="Run '/usr/sbin/pkg_info > patchlist.txt'"
	version_ins="Enter the FreeBSD relase in the format '6.1_15' for 'FreeBSD 6.1-RELEASE-p15'"
	patchline="Host/FreeBSD/pkg_info="
	versionline="Host/FreeBSD/release="
	patch="${args[0]}"
	version="${args[1]}"
	getFile "$patch_ins" "$patch"
	patch="$file"

	getString "$version_ins" "$version"
	version="$string"
	
	printCMD "$system" "$patch" "$version"
 	
	v=`echo "$version" | perl -pi -e 's/\n/\\\\n/g'`
	p=`cat "$patch" | perl -pi -e 's/\n/\\\\n/g'`

	echo "$fileheader$versionline$v" >> $kb
	echo "$fileheader$patchline$p" >> $kb
			
}

setCisco(){
	
	# Need to get one file and one string.
	
	# Host/Cisco/IOS
	# Host/Cisco/ASA
	# Host/Cisco/show_ver
	
	system=10
	system_name="Cisco"
	nessus_filematch="cisco-sa-"

	version_ins="Run 'show version'"
	versionline="Host/Cisco/show_ver="
	version="${args[0]}"
	getFile "$version_ins" "$version"
	version="$file"
	
	ios_line="Host/Cisco/IOS/Version="
	asa_line="Host/Cisco/ASA/Version="
	
	printCMD "$system" "$version"
 	
	v=`cat "$version" | perl -pi -e 's/\n/\\\\n/g'`

	echo "$fileheader$versionline$v" >> $kb
	
 	IOS=`cat ios_version | egrep '^.*IOS.*Version'`
 	ASA=`cat ios_version | egrep '^Cisco Adaptive Security Appliance Software Version'`

	if [ "$IOS" ] 
		then
			printf "[+] IOS Detected '%b'" "$IOS"
			echo "$fileheader$ios_line$IOS" >> $kb
	elif [ "$ASA" ]
		then
			printf "[+] ASA Detected '%b'" "$ASA"
			echo "$fileheader$asa_line$ASA" >> $kb
			
	else
			echo "[!] No Version infor Found, exciting"
			exit
			rm $kb
	fi
	
}



# setup defaults

fileheader='1323253913 1 '
localchecks="Host/local_checks_enabled=1"
nasl=`which nasl`
kb=`mktemp`
a=0
b=1
x=0
alloncmd=0

echo "$fileheader$localchecks" > $kb

# read in normal options from command line

while getopts "s:n:d:?" o
        do
                case "$o" in
                        s)     	system="$OPTARG";let a+=2;;
                        n)	nasl="$OPTARG";let a+=2;cmdnasl=1;;
                        d)     	nessus_dir="$OPTARG";let a+=2;cmddir=1;;                     
                        [?])   	printf "\nUsage: %b -s 'system type' [-d 'nessus plugin dir'] [-n 'loction of nasl'] optional argumants\n\nWhere system type is one of:\n`printSystems`\n\n** Entering no parameters will run this in wizard mode walking you thorough the data collection for your desired system\n" "$0" 
                               	exit 1;;
                esac
        done
        
# Parse all the other arguments into an array

for arg in "$@"
	do
		if [ $b -gt  $a ]
			then
				args[$x]="$arg"
				let x++
		fi
		let b++
	done

# checking to see if we have nessus nasl in our path

if [ ! -x "$nasl" ]
	then

		printf "[!] Cannot find nasl (%b), pass as -n argument\n" "$nasl" >&2
		exit 1;
fi

# checking to see if we have a Nessus Dir

if [ ! -d "$nessus_dir" ]
	then

		printf "[!] Cannot find Nessus plugin dir (%b), pass as -d argument\n" "$nessus_dir" >&2
		exit 1;
fi

# what system do you want?

while [ -z "$system_name" ]
	do
		case "$system" in 
			1) printf "[+] Redhat Selected\n">&2;setRedHat>&2;;
			2) printf "[+] OSX Selected\n">&2;setOSX>&2;;
			3) printf "[+] Debian Selected\n">&2;setDebian>&2;;
			4) printf "[+] Ubuntu Selected\n">&2;setUbuntu>&2;;
			5) printf "[+] Slackware Selected\n">&2;setSlackware>&2;;
			6) printf "[+] Solaris Selected\n">&2;setSolaris>&2;;
			7) printf "[+] AIX Selected\n">&2;setAIX>&2;;
			8) printf "[+] HP-UX Selected\n">&2;setHPUX>&2;;
			9) printf "[+] FreeBSD Selected\n">&2;setFreeBSD>&2;;
			10) printf "[+] Cisco Selected\n">&2;setCisco>&2;;
  			*) if [ ! -z "$system" ]; then printf "[!] Invalid System (%b) - try again\n" "$system" >&2; fi;getSystem >&2;;
		esac	
	done	

# find the nasls to use

printf "[+] Locating Nasls\n" >&2
nasls=`find "$nessus_dir" -iname "*$nessus_filematch*"` 

# run the checker on your kb file.

if [ -z "$nasls" ]
	then
		printf "[!] No Nasls found to test, check your Nessus dir is correct '%b'\n" "$nessus_dir"
		exit 1
fi

printf "[+] Checking for `echo \"$nasls\"| wc -l` Missing Patches\n" >&2
# The following section can be used to print the nasl that produced the output, slower as the run nasl everytime to get name that produced the ouput.

echo "$nasls" | while read a;
			do mp=`$nasl -k "$kb" "$a" | perl -pi -e 's/\n//' | perl -pi -e 's/Remote package/, Remote package/g' | perl -pi -e 's/Should be/, Should be/g' | perl -pi -e 's/Update to/, Update to/g' | perl -pi -e 's/Missing patch/, Missing patch/g'`; # run nasl
				if [ ! -z "$mp" ]; # check for any output from nasl 
					then printf "`basename $a .nasl`$mp\n"; # print name and output, this output is not very useful will need work
				fi;
			done;

#$nasl -k "$kb" $nasls ; # run nasl quickly, lists packages no nasls
			
# clean up 
 
 
#echo $kb 
rm $kb













